#ifndef LIGHT_H
#define LIGHT_H

#include <glm/vec3.hpp>
#include <glm/mat4x4.hpp>
#include <string>
#include <vector>

class FrameBuffer;
class Shader;
class Scene;
class Texture;
class Camera;

class Light
{
public:
    enum LightType {
        AMBIENT,
        DIRECTIONNAL,
        POINT,
        SPOT
    };

    enum LightFlags {
        AMBIENT_FLAG,
        DIRECTIONNAL_FLAG,
        POINT_FLAG,
        SPOT_FLAG,
        SHADOWMAP_FLAG,
        CASCADED_FLAG,
        NB_FLAGS
    };
    static const char* flagStr[];
    static const glm::mat4 biasMatrix;

    virtual LightType getType() = 0;

    glm::vec3 getColor() { return m_color; }

    virtual bool isShadowCaster() { return false; }

    virtual void bindAttributes(Shader *shader, Camera *camera) = 0;

    virtual unsigned int getFlags() = 0;

protected:
    glm::vec3 m_color;
};

class AmbientLight : public Light
{
public:
    AmbientLight(glm::vec3 lightColor = glm::vec3(0.1f)) { m_color = lightColor; }
    virtual LightType getType() { return AMBIENT; }
    virtual void bindAttributes(Shader *shader, Camera *camera);
    virtual unsigned int getFlags() { return 1 << AMBIENT_FLAG; }
};

class DirectionnalLight : public Light
{
public:
    DirectionnalLight(glm::vec3 dir = glm::vec3(0, -1, 0), glm::vec3 lightColor = glm::vec3(1));

    glm::vec3 getDir() { return m_direction; }
    
    virtual LightType getType() { return DIRECTIONNAL; }
    virtual void bindAttributes(Shader *shader, Camera *camera);
    virtual unsigned int getFlags();
    
    void initShadowMap(int resolution, bool isCascaded = false);
    void destroyShadowMap();
    void setShadowView(glm::vec3 dim = glm::vec3(1), glm::vec3 center = glm::vec3(0));
    void setCascadedShadowView(Camera *camera);
    void updateShadowMap(Scene* scene);
    
private:
    glm::vec3 m_direction;
    
    glm::vec3 m_center;
    
    bool m_shadowCaster;
    int m_shadowMapResolution;
    FrameBuffer* m_shadowMap;
    std::vector<glm::mat4> m_viewMatrices;
    std::vector<glm::mat4> m_projectionMatrices;
    
    static int m_shaderRefCounter;
    static Shader* m_shaders[4];
};

#endif // LIGHT_H
